#Copyright (c) 2017, SAS Institute Inc., Cary, NC, USA, All Rights Reserved

# 
# It was decided that gathering all output as a transcript better served the install
# and deployment needs so Write-Host was used rather than the more standard 
# Write-Output/Debug/Error/...
#
# These functions exist as wrappers so if it is decided to adopted mainline PS standards
# only these functions need to change.
#


<#
.Synopsis
Wrappers for logging support.

.Description
Simple functions to support consistent output of text messages.

#>

$ConsoleColors = @{
    ErrorForegroundColor    = [System.ConsoleColor]::Red;
    ErrorBackgroundColor    = [System.ConsoleColor]::Black;
    WarningForegroundColor  = [System.ConsoleColor]::Yellow;
    WarningBackgroundColor  = [System.ConsoleColor]::Black;
    DebugForegroundColor    = [System.ConsoleColor]::Yellow;
    DebugBackgroundColor    = [System.ConsoleColor]::Black;
    VerboseForegroundColor  = [System.ConsoleColor]::Yellow;
    VerboseBackgroundColor  = [System.ConsoleColor]::Black;
    ProgressForegroundColor = [System.ConsoleColor]::Yellow;
    ProgressBackgroundColor = [System.ConsoleColor]::DarkCyan
    }

$ForeColor = [System.ConsoleColor]::DarkYellow
$BackColor = [System.ConsoleColor]::DarkMagenta

$ghost = Get-Host
if ($ghost.PrivateData -ne $null) {
    $ConsoleColors = $ghost.PrivateData
    }

if ($ghost.UI -ne $null) {
    if ($ghost.UI.RawUI -ne $null) {
        $ForeColor = $ghost.UI.RawUI.ForegroundColor
        $BackColor = $ghost.UI.RawUI.BackgroundColor
        }
    }

#------------------------------------------------------------------------------------------    
function FormatLogMessage([string] $lvl, [string] $msg) {
    return "{0} {1} - {2}" -f (get-date -format "yyyyMMdd HH:mm:ss:fffff"), $lvl, $msg
    }


<#
.Synopsis
Write message to output log with no ornamentation (datetime, type prefix)

.Description
Writes the given message to the output log, sans ornamentation.

.Parameter msg
String message to output

.Outputs
None
#>
function Write-Log {

    [CmdletBinding()]
    param ([string] $msg = "")

    Write-Host -BackgroundColor $BackColor `
               -ForegroundColor $ForeColor `
               ($msg)
    }


<#
.Synopsis
Write message to debug output log

.Description
Writes the given message to the debug output log

.Parameter msg
String message to output

.Outputs
None
#>
function Write-LogDebug {
    
    [CmdletBinding()]
    param ([string] $msg)
    
    $pref = $PSCmdlet.GetVariableValue('DebugPreference')
    if ($pref -ne "SilentlyContinue") {
        Write-Host -BackgroundColor $ConsoleColors.DebugBackgroundColor `
                   -ForegroundColor $ConsoleColors.DebugForegroundColor `
                   (FormatLogMessage "Debug " $msg)
        }
    }

<#    
.Synopsis
Write an error message to output log.

.Description
Writes the given message, category, exception or error record to the error output log.

.Parameter msg
String message to output

.Parameter msg
String message to output

.Parameter msg
String message to output

.Parameter msg
String message to output

.Outputs
None
#>
function Write-LogError {

    [CmdletBinding()]
    param (
        [Parameter(ParameterSetName="MsgArg", Position="0")]
        [string] $msg,
        [Parameter(ParameterSetName="MsgArg", Mandatory=$false)]
        [System.Management.Automation.ErrorCategory] $category = ([System.Management.Automation.ErrorCategory]::NotSpecified),

        [Parameter(ParameterSetName="ExceptionArg", Position="0")]
        [System.Exception] $exception = $null,

        [Parameter(ParameterSetName="ErrorRecArg", Position="0")]                        
        [System.Management.Automation.ErrorRecord] $errorRecord = $null
        )
        
    $tb = $null
    $cat = $null
    switch ($PsCmdlet.ParameterSetName) {
        "MsgArg" {
            if (($category -ne $null) -and 
                ($category -ne [System.Management.Automation.ErrorCategory]::NotSpecified)) {
                $cat = $category
                }
            }

        "ExceptionArg" {
            if ($exception -ne $null) {
                $msg = $exception.Message
                $tb = $exception.StackTrace
                }
            }

        "ErrorRecArg" {
            if ($errorRecord -ne $null) {
                if ([String]::IsNullOrWhiteSpace($errorRecord.PSMessageDetails) -eq $false) {
                    $msg = $errorRecord.PSMessageDetails            
                    }
                elseif ([String]::IsNullOrWhiteSpace($errorRecordException.Message) -eq $false) {
                    $msg = $errorRecord.Exception.Message              
                    }
                elseif ($errorRecord.Exception -ne $null) {
                    $msg = $errorRecord.Exception.Message
                    }
                $tb = $errorRecord.ScriptStackTrace
                $cat = $errorRecord.CategoryInfo.Category
                }
            }

        }

    if ($cat -ne $null) {
        $msg = "[$cat]" + $msg
        }
        
    if ($tb -ne $null) {
        $msg += "`r`n" + $tb
        }
    
    $pref = $PSCmdlet.GetVariableValue('ErrorActionPreference')
#    Write-Host -ForegroundColor Green "Error pref: $pref"
    if ($pref -ne "SilentlyContinue") {
        Write-Host -BackgroundColor $ConsoleColors.ErrorBackgroundColor `
                   -ForegroundColor $ConsoleColors.ErrorForegroundColor `
                   (FormatLogMessage "ERROR" $msg)
        }
    }    


<#    
.Synopsis
Write an infomational message to output log.

.Description
Writes the given message to the information output log.

.Parameter msg
String message to output

.Outputs
None
#>
function Write-LogInfo {
    
    [CmdletBinding()]
    param ([string] $msg)
    
    $pref = $PSCmdlet.GetVariableValue('$InformationPreference')
    if ($pref -ne "SilentlyContinue") {
        Write-Host  -BackgroundColor $ConsoleColors.ProgressBackgroundColor `
                    -ForegroundColor $ConsoleColors.ProgressForegroundColor `
                   (FormatLogMessage "Info " $msg)
        }
    }


<#    
.Synopsis
Write a warning verbose to output log.

.Description
Writes the given message to the verbose output log.

.Parameter msg
String message to output

.Outputs
None
#>
function Write-LogVerbose {
    
    [CmdletBinding()]
    param ([string] $msg)
    
    $pref = $PSCmdlet.GetVariableValue('VerbosePreference')
    if ($pref -ne "SilentlyContinue") {
        Write-Host -BackgroundColor $ConsoleColors.VerboseBackgroundColor `
                   -ForegroundColor $ConsoleColors.VerboseForegroundColor `
                   (FormatLogMessage "Warn " $msg)
        }
    }


<#    
.Synopsis
Write a warning message to output log.

.Description
Writes the given message to the warning output log.

.Parameter msg
String message to output

.Outputs
None
#>
function Write-LogWarning {
    
    [CmdletBinding()]
    param ([string] $msg)
    
    $pref = $PSCmdlet.GetVariableValue('WarningPreference')
    if ($pref -ne "SilentlyContinue") {
        Write-Host -BackgroundColor $ConsoleColors.WarningBackgroundColor `
                   -ForegroundColor $ConsoleColors.WarningForegroundColor `
                   (FormatLogMessage "Warn " $msg)
        }
    }
