<#
.SYNOPSIS
Given various install data, run a configuration for products that
support it.

.DESCRIPTION
Long description

.PARAMETER productsToConfigure
A list of products that can be configured

.PARAMETER pkgMetadata
A list of metadata objects for all products being installed.

.PARAMETER installDir
Directory that the MSIs being configured were installed to

.PARAMETER licenseFile
Path to the license file.

.PARAMETER programDataConfig
Path to the configuration directory under ProgramData

.PARAMETER publicConfigDir
Path to the publicly accessible config files

.PARAMETER instanceId
An instance ID for your deployment

.PARAMETER deploymentId
Your deployment ID

.EXAMPLE
Publish-Configuration -productsToConfigure @('prod1','prod2') -pkgMetadata $pkgMetadata -installDir "C:\Program Files\SAS" `
    -licenseFile "C:\some\license\file.txt" -programDataConfig "C:\ProgramData\sas" -publicConfigDir "C:\Users\Public\config" `
    -instanceId "default" -deploymentId "Viya"
#>
function Publish-Configuration {
    [CmdletBinding()]
    param(
        [array] $productsToConfigure,
        [psobject] $pkgMetadata,
        [string] $installDir,
        [string] $licenseFile,
        [string] $programDataConfig,
        [string] $publicConfigDir,
        [string] $instanceId,
        [string] $deploymentId
    )

    $deploymentData = New-Object -TypeName psobject -Property @{
        installDir           = $installDir
        licenseFile          = $licenseFile
        programDataConfigDir = $programDataConfig
        publicConfigDir      = $publicConfigDir
        instanceId           = $instanceId
        deploymentId         = $deploymentId
    }
    $configData = [ConfigData]::new($pkgMetadata, $deploymentData)

    foreach ($product in $productsToConfigure) {
        Write-Debug "product to configure: $product"
        RunConfiguration $product $configData
    }
}


<#
.SYNOPSIS
Removes an existing configuration

.DESCRIPTION
Removes the configurations of products listed by scanning the uninstall data that was
laid down at install time.

.PARAMETER pkgMetadata
Hashtable of data about the installed products

.PARAMETER productsToConfigure
List of configurable products

.PARAMETER installDir
Directory that the MSIs being configured were installed to

.PARAMETER licenseFile
Path to the license file.

.PARAMETER programDataConfig
Path to the configuration directory under ProgramData

.PARAMETER publicConfigDir
Path to the publicly accessible config files

.PARAMETER instanceId
An instance ID for your deployment

.PARAMETER deploymentId
Your deployment ID

.EXAMPLE
Remove-Configuration -productsToConfigure @('prod1','prod2') -pkgMetadata $pkgMetadata -installDir "C:\Program Files\SAS" `
    -licenseFile "C:\some\license\file.txt" -programDataConfig "C:\ProgramData\sas" -publicConfigDir "C:\Users\Public\config" `
    -instanceId "default" -deploymentId "Viya"
#>
function Remove-Configuration {
    [CmdletBinding()]
    param (
        [array] $productsToConfigure, #todo: these need to be part of the pkgMetadata
        [psobject] $pkgMetadata,
        [string] $installDir,
        [string] $licenseFile,
        [string] $programDataConfig,
        [string] $publicConfigDir,
        [string] $instanceId,
        [string] $deploymentId
    )

    $deploymentData = New-Object -TypeName psobject -Property @{
        installDir           = $installDir
        licenseFile          = $licenseFile
        programDataConfigDir = $programDataConfig
        publicConfigDir      = $publicConfigDir
        instanceId           = $instanceId
        deploymentId         = $deploymentId
    }
    $configData = [ConfigData]::new($pkgMetadata, $deploymentData)

    foreach ($product in $productsToConfigure) {
        Write-Debug "product to configure: $product"
        RunUnconfiguration $product $configData
    }
}

# TODO: make sure product is deployed in the environment. Specifically, we
# want it in a function. Check for the uninstall reg keys.

<#
.SYNOPSIS
Reads the config file to determine the configurable products

.DESCRIPTION
Reads config-start-sequence.psd1 to determine the products to be configured,
and compares them with the products to be installed. Whatever exists in both
sets will be configured.

.PARAMETER entitlementData
An array of psobjects containing data about the products being installed

.PARAMETER root
The root directory containing the configuration directory.

.EXAMPLE
Get-ConfigurableProducts -entitlementData @(@{name="prodA"}, @{name="prodB"}) -root $PSScriptRoot

#>
function Get-ConfigurableProducts {
    [CmdletBinding()]
    param(
        [psobject[]] $entitlementData,
        [string] $root
    )
    $configDir = Join-Path $root "config"
    $products = Import-LocalizedData -BaseDirectory $configDir -FileName "config-start-sequence.psd1"
    $twelveBytes = $entitlementData.TwelveByte
    $configurables = $products.sequence | Where-Object {$twelveBytes.Contains($_)}
    return $configurables
}
