function GetPackageMetadata {
    param(
        [Parameter(Mandatory = $true)]
        [string] $repoUrl,
        [Parameter(Mandatory = $true)]
        [string] $caPath,
        [Parameter(Mandatory = $true)]
        [string] $certPath,
        [Parameter(Mandatory = $true)]
        [string] $downloadDir,
        [switch] $forceRefresh
    )

    $cachePath = Join-Path $downloadDir ".metadataCache.xml"
    $cacheExists = Test-Path $cachePath
    if ($cacheExists -and !($forceRefresh)) {
        Write-Verbose "Using cached metadata at '$cachePath'"
        $pkgList = New-Object -TypeName System.Collections.ArrayList
        $cacheMd = Import-UninstallData -uninstallDataFile $cachePath

        #cacheMd is imported as a hashtable, so we need to convert it back to a
        #powershell object.
        foreach ($entry in $cacheMd.GetEnumerator()) {
            $pkg = New-Object -TypeName psobject -Property $entry.Value
            $pkgList.Add($pkg) | Out-Null
        }

        return $pkgList
    }

    $repomd = Get-EntitlementMetadata -baseurl $repoUrl -rootCAPath $caPath `
        -entitlementCertPath $certPath -Verbose:$VerbosePreference
    Write-Verbose "Caching package metadata at '$cachePath'"
    Save-UninstallData $repomd $cachePath -Verbose:$VerbosePreference | Out-Null
    return $repomd
}

#Compares installed products with repo equivalents, and retuns a list of products to update.
function GetProductsToInstallAndUpdate {
    param(
        [psobject[]] $pkgMetadata
    )

    $productsToUpdate = New-Object -TypeName System.Collections.ArrayList
    $productsToInstall = New-Object -TypeName System.Collections.ArrayList
    $installedProducts = @(Get-InstalledProducts)

    foreach ($product in $pkgMetadata) {
        $installed = $installedProducts | Where-Object {$_.ProductName -eq $product.Name}

        if ($null -ne $installed) {
            $verComp = Compare-Version $installed.Version $product.Version

            #only add to updates if newer version is available.
            if ($verComp -lt 0) {
                $productsToUpdate.Add($product) | Out-Null
            }
        }
        else {
            $productsToInstall.Add($product) | Out-Null
        }
    }

    $installChains = New-Object -TypeName psobject -Property @{
        ProductsToInstall = $productsToInstall
        ProductsToUpdate  = $productsToUpdate
    }

    return $installChains
}

function GetProductsToDownload {
    param(
        [psobject[]] $pkgMetadata, #from sas
        [string] $entCertPath,
        [string] $rootCaPath,
        [switch] $checkAllVersionLevels
    )

    $pkgUrls = New-Object -TypeName System.Collections.ArrayList
    $installedProducts = @(Get-InstalledProducts) #retrieves products installed from registry
    $updateProducts = @(getProductToUpdate $pkgMetadata -installedProducts $installedProducts `
            -entCertPath $entCertPath -rootCAPath $rootCaPath -checkAllVersionLevels:$checkAllVersionLevels)
    $newProducts = @(getProductsToInstall $pkgMetadata -installedProducts $installedProducts)

    if ($null -ne $updateProducts) {
        $pkgUrls.AddRange($updateProducts)
    }

    if ($null -ne $newProducts) {
        $pkgUrls.AddRange($newProducts)
    }

    return $pkgUrls.ToArray()
}

function getProductToUpdate {
    param (
        [psobject[]] $pkgMetadata,
        [psobject[]] $installedProducts,
        [string] $entCertPath,
        [string] $rootCaPath,
        [switch] $checkAllVersionLevels
    )

    $installedProductsHash = @{}
    $productUpdates = New-Object -TypeName System.Collections.ArrayList
    $updatesAvailableFromSas = $null

    if ($null -ne $installedProducts -and $installedProducts.Count -gt 0) {

        foreach ($product in $installedProducts) {
            $installedProductsHash.add($product.ProductName, $product.Version) #data is 4 levels Maj.Min.Bld.Rev by default
        }
        #get arraylist of updates available based on version level of 3 or 4
        $updatesAvailableFromSas = Show-Update -products $installedProductsHash -entitlementCertPath $entCertPath `
            -rootCAPath $rootCaPath -checkAllVersions:$checkAllVersionLevels
    }

    foreach ($product in $pkgMetadata) {
        if ($null -ne $updatesAvailableFromSAS ) {
            $prod = $updatesAvailableFromSas | Where-Object {$_.Name -eq $product.Name}
            if ($null -ne $prod) {
                $productUpdates.Add($product) | Out-Null
            }
        }
    }

    return $productUpdates
}

function getProductsToInstall {
    param (
        [psobject[]] $pkgMetadata,
        [psobject[]] $installedProducts
    )

    $newProducts = New-Object -TypeName System.Collections.ArrayList

    foreach ($product in $pkgMetadata) {
        $prod = $installedProducts | Where-Object {$_.ProductName -eq $product.Name}

        if ($null -eq $prod) {
            $newProducts.Add($product) | Out-Null
        }
    }

    return $newProducts
}