
<#
.SYNOPSIS
Fetches and installs sas given the install data provided.

.DESCRIPTION
Given a zip file containing entitlement data for an order, Install-Sas
will fetch and install all packages that a user is entitled to

.PARAMETER download
If specified, download files for install

.PARAMETER install
If specified, run the install

.PARAMETER config
If specified run the config

.PARAMETER version
If specified display the version of the deployment tool

.PARAMETER deployDataFile
Path to the zip file containing the deployment data. If empty, it
will search in the root of the ViyaWindowsDeployment project.

.PARAMETER installDir
Directory to install the fetched packages to

.PARAMETER downloadDir
Directory to store downloaded packages in.

.EXAMPLE
Install-Sas

.NOTES
If neither -install or -config are specified, both will be run
#>
function Install-Sas() {
    [CmdletBinding()]
    param (
        [switch] $download = $false,
        [switch] $install = $false,
        [switch] $config = $false,
        [switch] $version = $false,
        [string] $deployDataFile,
        [string] $installDir = $(Join-Path $env:ProgramFiles "SAS"),
        [string] $downloadDir,
        [switch] $checkAllVersions,
        [switch] $noPrompt
    )

    Set-StrictMode -Version 5

    #Set directory tracking structure
    $staticFile = Join-Path $PSScriptRoot "static_vars.psd1"
    $staticVars = Import-PowerShellDataFile $staticFile
    $statics = [StaticVariables]::new($staticVars)
    $hierarchy = [ViyaHierarchy]::new($installDir, $statics.deploymentId, $PSScriptRoot)

    # If version switch is passed in, display version and then exit tool.
    if ($version.IsPresent) {
        $statics.version
        exit 0
    }

    if ($null -eq $downloadDir -or '' -eq $downloadDir) {
        $downloadDir = $(Join-Path $hierarchy.GetDeploymentToolsRoot() "Downloads")
    }

    $timestamp = Get-Date -Format FileDateTime
    $logFilename = "deploy-$timestamp.log"
    $logFile = Join-Path $hierarchy.GetDeploymentToolsRoot() $logFilename
    Start-Transcript -Path $logFile

    # If neither install, config, nor download are selected, default to performing the actions download, install, and config
    if ($download -eq $false -and $install -eq $false -and $config -eq $false) {
        $download = $true
        $install = $true
        $config = $true
    }

    $date = Get-Date
    Write-Host "Deployment started at $date"

    ImportRequiredModules $hierarchy.GetDeploymentModuleLibrary()

    if ($null -eq $deployDataFile -or '' -eq $deployDataFile) {
        $deployDataFile = Join-Path $hierarchy.GetDeploymentToolsRoot() "SAS_Viya_deployment_data.zip"
    }

    # define SAS Deployment Data content paths
    $entCertDir = Join-Path $hierarchy.GetDeploymentToolsRoot() "entitlement-certificates"
    $caCertDir = Join-Path $hierarchy.GetDeploymentToolsRoot() "ca-certificates"
    $licenseDir = Join-Path $hierarchy.GetDeploymentToolsRoot() "licenses"

    # Unzip the contents of the SAS Deployment Data zip to the script root directory
    # if the full path of SAS_Viya_Deployment_Data.zip exists. The old content will
    # be deleted prior to unzipping the new content.
    if (Test-Path -Path $deployDataFile) {
        if (Test-Path $licenseDir) { Remove-Item $licenseDir -Recurse; }
        if (Test-Path $caCertDir) { Remove-Item $caCertDir -Recurse; }
        if (Test-Path $entCertDir) { Remove-Item $entCertDir -Recurse; }
        Expand-Archive $deployDataFile -DestinationPath $hierarchy.GetDeploymentToolsRoot() -Force
    }

    # verify license paths
    if (!(Test-Path -Path $licenseDir)) {
        Write-Error "Deployment data for your order cannot be found.`nERROR: The 'licenses' directory does not exist."
        exit 1
    }
    $licenses = @(Get-ChildItem $licenseDir)
    if ($licenses.Count -eq 0) {
        Write-Error "Deployment data for your order cannot be found.`nERROR: A license file must exist in the 'licenses' directory."
        exit 1
    }
    $licenseFile = $licenses | Select-Object -First 1
    if ($licenses.Count -gt 1) {
        Write-Warning "More than one license file found at '$licenseDir', using license file '$licenseFile' "
    }
    $licensePath = Join-Path $licenseDir $licenseFile
    Write-Host "Using license file: $licensePath"

    # define and verify certificate paths
    $caCertPath = Join-Path $caCertDir $statics.caCert
    if (!(Test-Path -Path $caCertPath)) {
        Write-Error "Deployment data for your order cannot be found.`nERROR: The '$caCertPath' file does not exist."
        exit 1
    }
    $entCertPath = Join-Path $entCertDir $statics.entitlementCert
    if (!(Test-Path -Path $entCertPath)) {
        Write-Error "Deployment data for your order cannot be found.`nERROR: The '$entCertPath' file does not exist."
        exit 1
    }

    $pkgMetadata = $null

    if ($download) {
        if ($null -eq $pkgMetadata) {
            $pkgMetadata = GetPackageMetadata -repoUrl $statics.repositoryUrl -caPath $caCertPath -certPath $entCertPath `
                -downloadDir $downloadDir -forceRefresh
        }

        $products = GetProductsToDownload -pkgMetadata $pkgMetadata -entCertPath $entCertPath `
            -rootCAPath $caCertPath -checkAllVersionLevels:$checkAllVersions

        foreach ($product in $products) {
            Get-EntitlementByUrl -downloadDir $downloadDir -entitlementCertPath $entCertPath `
                -rootCAPath $caCertPath -url $product.Url -checksum $product.checksum -checksumType $product.checksumType `
                -Verbose:$VerbosePreference
        }
    }

    if ($install) {
        if ($null -eq $pkgMetadata) {
            $pkgMetadata = GetPackageMetadata -repoUrl $statics.repositoryUrl -caPath $caCertPath -certPath $entCertPath -downloadDir $downloadDir
        }

        $installList = GetProductsToInstallAndUpdate $pkgMetadata
        $productsToInstall = $installList.ProductsToInstall
        $productsToUpdate = $installList.ProductsToUpdate

        if ($productsToUpdate.Count -gt 0) {
            if ($true -eq $noprompt){
                $answer = $true
            }
            else {
                $answer = PromptInstallUpgrade
            }
            if ($true -eq $answer) {
                Install-SasPackage $productsToUpdate $downloadDir $installDir $hierarchy.GetInstallLogDir() -Verbose:$VerbosePreference
                Save-UninstallData $productsToUpdate $hierarchy.GetUninstallDataFile() -Verbose:$VerbosePreference
            }
            else {
                #If the user decides not to upgrdade, terminate the script.
                Write-Host "No install will be performed."
                TerminateEarly
            }
        }

        # distribute deployment modules to the system.
        Install-DeploymentModules -sourceDir $hierarchy.GetDeploymentToolsRoot() -destinationDir $hierarchy.GetModulesDir() -Verbose:$VerbosePreference
        Install-SasPackage $productsToInstall $downloadDir $installDir $hierarchy.GetInstallLogDir() -Verbose:$VerbosePreference
        Save-UninstallData $productsToInstall $hierarchy.GetUninstallDataFile() -Verbose:$VerbosePreference
    }

    if ($config) {
        $installMd = Import-UninstallData $hierarchy.GetUninstallDataFile() -Verbose:$VerbosePreference
        $configProducts = Get-ConfigurableProducts -entitlementData $installMd.Values -root $hierarchy.GetDeploymentToolsRoot() -Verbose:$VerbosePreference
        Publish-Configuration $configProducts $installMd $installDir $licensePath `
            $hierarchy.GetConfigDir() $hierarchy.GetPublicSasDir() $statics.instanceId $statics.deploymentId -Verbose:$VerbosePreference
    }

    $date = Get-Date
    Write-Host "Deployment completed at $date"

    Stop-Transcript
}

<#
.SYNOPSIS
Unconfigures and then uninstalls installed software

.DESCRIPTION
Scans the install directory for uninstall metadata, and then run unconfigure
where needed, and uninstall on all packages. If neither -uninstall nor -config
are specified, both uninstall and unconfigure will be run.

.PARAMETER uninstall
Flag to uninstall. Defaults to false

.PARAMETER config
Flag to configure. Defaults to false.

.PARAMETER installDir
Path to the install directory

.EXAMPLE
Remove-Sas

.NOTES
Will delete everything living in download directory.
#>
function Remove-Sas {
    [CmdletBinding()]
    param (
        [switch] $uninstall = $false,
        [switch] $config = $false,
        [string] $installDir = $(Join-Path $env:ProgramFiles "SAS")
    )

    # If neither install or config is selected, default to performing the actions install and config
    if ($uninstall -eq $false -and $config -eq $false) {
        $uninstall = $true
        $config = $true
    }

    Set-StrictMode -Version 5

    #Set directory tracking structure
    $staticFile = Join-Path $PSScriptRoot "static_vars.psd1"
    $staticVars = Import-PowerShellDataFile $staticFile
    $statics = [StaticVariables]::new($staticVars)
    $hierarchy = [ViyaHierarchy]::new($installDir, $statics.deploymentId, $PSScriptRoot)

    $timestamp = Get-Date -Format FileDateTime
    $logFilename = "remove-$timestamp.log"
    $logFile = Join-Path $hierarchy.GetDeploymentToolsRoot() $logFilename
    Start-Transcript -Path $logFile

    ImportRequiredModules $hierarchy.GetDeploymentModuleLibrary()

    $uninstallData = Import-UninstallData $hierarchy.GetUninstallDataFile() -Verbose:$VerbosePreference

    if ($config) {
        $configProducts = Get-ConfigurableProducts -entitlementData $uninstallData.Values -root $hierarchy.GetDeploymentToolsRoot() -Verbose:$VerbosePreference
        Remove-Configuration $configProducts $uninstallData $installDir "nolicense" `
            $hierarchy.GetConfigDir() $hierarchy.GetPublicSasDir() $statics.instanceId $statics.deploymentId -Verbose:$VerbosePreference
    }

    if ($uninstall) {
        Write-Host "Uninstalling SAS software"
        Uninstall-SasPackage $installdir $hierarchy.GetInstallLogDir() $hierarchy.GetUninstallDataFile() -Verbose:$VerbosePreference
    }

    Stop-Transcript
}

function ImportRequiredModules {
    param(
        [string] $moduleLib
    )

    Import-Module -FullyQualifiedName "$moduleLib\SASPsd1FileManagement" -Force
    Import-Module -FullyQualifiedName "$moduleLib\SasInstaller" -Force
    Import-Module -FullyQualifiedName "$moduleLib\SasConfiguration" -Force
    Import-Module -FullyQualifiedName "$moduleLib\SasEntitlement" -Force
}
