Class RepoController {
    [OddData] $odd
    #repos are composed of (repopath/Repository)
    [hashtable] $metadataRepos = @{}
    [hashtable] $productRepos = @{}

    RepoController([hashtable] $repos, [System.Security.Cryptography.X509Certificates.X509Certificate2] $entitlementCert) {

        foreach ($entry in $repos.GetEnumerator()) {
            if ($entry.Key.StartsWith('sasmd/')) {
                $this.metadataRepos[$entry.Key] = $entry.Value
                continue
            }

            $this.productRepos[$entry.Key] = $entry.Value
        }

        $urls = @($this.GetOddFileUrls())
        $this.CompileOddData($urls, $entitlementCert)
    }

    [Repository[]] GetMetadataRepos() {
        return $this.metadataRepos.Values
    }

    [Repository[]] GetProductRepos() {
        return $this.productRepos.Values
    }

    [string[]] GetOddFileUrls() {
        $oddFileUrls = @()

        foreach ($repo in $this.metadataRepos.GetEnumerator()) {
            $oddFileUrls += @($repo.Value.GetOddFilesUrls())
        }

        if ($null -eq $oddFileUrls -or $oddFileUrls.Count -eq 0) {
            throw "No odd files could be found in the metadata repos."
        }

        return $oddFileUrls
    }

    [void] CompileOddData([string[]] $oddFileUrls, [System.Security.Cryptography.X509Certificates.X509Certificate2] $entitlementCert) {
        $this.odd = [OddData]::new()

        foreach ($url in $oddFileUrls) {
            $oddData = DownloadControlData $url $entitlementCert
            $dus = @($oddData.deployableUnits)
            $this.odd.Merge($dus)
        }
    }

    [PSCustomObject] GetPackageData([string] $name) {
        $twelveByte = $this.odd.GetTwelveByteByUnitName($name)
        $pkgData = $This.GetLatestPackageInRepos($twelveByte , $name)

        return $pkgData
    }

    [PSCustomObject] GetLatestPackageInRepos([string] $twelveByte, [string] $pkgName) {
        $latestPkgData = $null

        foreach ($entry in $this.productRepos.GetEnumerator()) {
            $pkg = $entry.Value.GetPackageByName($pkgName)
            $url = $entry.Value.GetPackageUrlByName($pkgName)

            if ($null -eq $latestPkgData) {
                $latestPkgData = $this.newPackageData($twelveByte, $pkg, $url)
                continue
            }

            #if $pkg is newer
            $comp = compareVersions -first $latestPkgData.Version -second $pkg.Version
            if ($comp -lt 0) {
                $latestPkgData = $this.newPackageData($twelveByte, $pkg, $url)
            }
        }

        if ($null -eq $latestPkgData) {
            Write-Error "no repo contains '$pkgName'"
        }

        return $latestPkgData
    }

    hidden [PSCustomObject] newPackageData([string] $twelveByte, [Package] $pkg, [string] $url ) {

        $pkgData = New-Object -TypeName psobject -Property @{
            Package    = $pkg
            Url        = $url
            TwelveByte = $twelveByte
        }

        return $pkgData
    }
}

Class Repository {
    [string] $baseUrl
    [string] $subUrl
    [RepoMd] $metadata
    [Primary] $primary

    Repository([string] $baseurl, [string] $subUrl, [RepoMd] $metadata, [Primary] $primary) {
        $this.baseUrl = $baseurl
        $this.subUrl = $subUrl
        $this.metadata = $metadata
        $this.primary = $primary
    }

    #Always takes latest
    [Package] GetPackageByName([string] $name) {
        return $this.primary.GetLatestPackage($name)
    }

    [Package] GetPackageByNameAndProductCode([string] $name, [string] $productCode) {
        return $this.primary.GetPackageByNameAndProductCode($name, $productCode)
    }

    [string] GetPackageUrlByName([string] $name) {
        $repourl = JoinUrlPieces @($this.baseUrl, $this.subUrl)
        $pkg = $this.GetPackageByName($name)
        $pkgUrl = JoinUrlPieces @($repourl, $pkg.location)

        return $pkgUrl
    }

    [string[]] GetOddFilesUrls() {
        $urls = New-Object -TypeName System.Collections.ArrayList
        $pkgs = @($this.primary.GetOddFiles())

        foreach ($pkg in $pkgs) {
            $url = JoinUrlPieces @($this.baseUrl, $this.subUrl, $pkg.location)
            $urls.Add($url) | Out-Null
        }

        return $urls
    }
}

Class RepoMd {
    [RepoMdData] $primaryData

    RepoMd([xml] $repodata) {
        $primary = @($repodata.repomd.data).Where{$_.type -eq "primary"}
        $this.primaryData = [RepoMdData]::new($primary.type, $primary.location.href, $primary.size,
            $primary.checksumType, $primary.checksum, $primary.timestamp)
    }

    [string] GetPrimaryFileChecksumType() {
        return $this.primaryData.checksumType
    }

    [string] GetPrimaryFileChecksum() {
        return $this.primaryData.checksum
    }

    [int] GetPrimaryFileSize() {
        return $this.primaryData.size
    }

    [string] GetPrimaryFileLocation() {
        return $this.primaryData.location
    }
}

Class RepoMdData {
    [string] $type
    [string] $location
    [string] $timestamp
    [string] $size
    [string] $checksumType
    [string] $checksum

    RepoMdData([string] $type, [string] $location, [string] $size, [string] $checksumType,
        [string] $checksum, [string] $timestamp) {
        $this.type = $type
        $this.location = $location
        $this.size = $size
        $this.checksum = $checksum
        $this.checksumType = $checksumType
        $this.timestamp = $timestamp
    }
}

Class Primary {
    [System.Array] $packages

    Primary([xml] $primary) {
        $this.packages = @()

        foreach ($package in @($primary.metadata.package)) {
            $pkg = [Package]::new($package)
            $this.packages += $pkg
        }
    }

    [Package] GetLatestPackage([string] $name) {
        $pkgMatch = $this.packages.Where{$_.name -eq $name}
        $versions = @($pkgMatch.version)
        $netVersions = $versions | ForEach-Object {[System.Version]$_}
        $latest = $netVersions | Sort-Object -Descending | Select-Object -First 1
        $latestPkg = $pkgMatch.Where{$_.version -eq $latest}

        return $latestPkg[0]
    }

    [Package] GetPackageByNameAndProductCode([string] $name, [string] $productCode) {
        $pkg = $this.packages.Where{$_.name -eq $name -and $_.productCode -eq $productCode}
        return $pkg[0]
    }

    [Package[]] GetOddFiles() {
        $pkg = $this.packages.Where{$_.type -eq 'odd'}
        return $pkg
    }
}

Class Package {
    [string] $type
    [string] $productCode
    [string] $upgradeCode
    [string] $productName
    [string] $checksum
    [string] $checksumType
    [string] $name
    [string] $version
    [string] $location
    [int] $size

    Package([Xml.XmlElement] $package) {
        $this.type = $package.type
        $this.productCode = $package.ProductCode
        $this.upgradeCode = $package.UpgradeCode
        $this.name = $package.Name
        $this.productName = $package.ProductName
        $this.checksum = $package.checksum.'#text'
        $this.checksumType = $package.checksum.type
        $this.version = $package.version
        $this.location = $package.location.href
    }
}

Class OddData {
    [System.Array] $units

    OddData() {
        $this.units = @()
    }

    #Takes information parsed from odd files and merges it with
    #what's already in the object
    [void] Merge([PSCustomObject] $data) {
        foreach ($entry in $data) {
            $unit = [Unit]::new($entry.unitName, $entry.name, $entry.unitRef,
                $entry.unitType)
            $this.units += $unit
        }
    }

    [System.Array] GetProducts() {
        $unitNames = @()
        foreach ($unit in $this.units) {
            $unitNames += $unit.unitName
        }

        return $unitNames | Select-Object -Unique
    }

    [System.Array] GetRepositories() {
        $repoNames = @()
        foreach ($unit in $this.units) {
            $repoNames += $unit.repository
        }

        return $repoNames | Select-Object -Unique
    }

    [string] GetPackageRepositoryPath([string] $packageName) {
        foreach ($unit in $this.units) {
            if ($packageName -eq $unit.unitName) {
                return $unit.repository
            }
        }

        throw "No repo found for package '$packageName'"
    }

    [string] GetTwelveByteByUnitName([string] $unitName) {
        foreach ($unit in $this.units) {
            if ($unitName -eq $unit.unitName) {
                return $unit.name
            }
        }

        throw "No twelve byte found for unit name '$unitName'"
    }
}

Class Unit {
    [string] $unitName
    [string] $name #name actually corresponds to the 12-byte code of the product
    [string] $repository
    [string] $unitType

    Unit([string] $unitName, [string] $name, [string] $repository,
        [string] $unitType) {
        $this.unitName = $unitName
        $this.name = $name
        $this.repository = $repository
        $this.unitType = $unitType
    }
}
