#This class should encapsulate all request processing so we have an easier time mocking.
#or this would be a class, but Pester doesn't support mocking objects yet...
function DownloadControlData() {
    param(
        [string] $url,
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $entitlementCert
    )
    Write-Verbose "Invoke-WebRequest: DownloadControlData: uri='$url'"
    $results = Invoke-WebRequest -Uri $url -Certificate $entitlementCert -UseBasicParsing
    Write-Verbose "Entitlement data returned from the web request:"
    Write-Verbose $results
    $jsonResponse = ConvertFrom-Json $results

    return $jsonResponse
}

function DownloadRepoMD {
    param(
        [string] $url,
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $entitlementCert
    )

    Write-Verbose "Invoke-WebRequest: DownloadRepoMD: uri='$url'"
    [xml]$results = Invoke-WebRequest -Uri $url -Certificate $entitlementCert -UseBasicParsing
    return $results
}

function DownloadPrimary {
    param(
        [string] $primaryFileUrl,
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $entitlementCert,
        [string] $primaryFileGz = (Join-Path $env:TEMP "primary.xml.gz")
    )

    Write-Verbose $primaryFileUrl
    Write-Verbose "Compiling package urls..."
    $outfile = $primaryFileGz.Replace(".gz", "")
    Invoke-WebRequest -Uri $primaryFileUrl -Certificate $entitlementCert -OutFile $primaryFileGz -UseBasicParsing
    Write-Verbose "Invoke-WebRequest: DownloadPrimary: uri='$primaryFileUrl'"
    $input = $null
    $output = $null
    $gzipStream = $null

    try{
        #Unzip the primary.xml file
        $input = New-Object System.IO.FileStream $primaryFileGz, ([System.IO.FileMode]::Open), ([System.IO.FileAccess]::Read)
        $output = New-Object System.IO.FileStream $outFile, ([System.IO.FileMode]::Create), ([System.IO.FileAccess]::Write)
        $gzipStream = New-Object System.IO.Compression.GZipStream $input, ([System.IO.Compression.CompressionMode]::Decompress)
        $bufferSize = 1024
        $buffer = New-Object Byte[] $bufferSize

        while(($bytesRead = $gzipStream.Read($buffer, 0, $bufferSize)) -ne 0){
            $output.Write($buffer, 0, $bytesRead)
        }

        $output.Flush()
    } catch {
        $errorMessage = $_.Exception.Message
        throw "An error occurred retrieving the primary XML: '$errorMessage'"
    } finally {
        if($null -ne $gzipStream){
            $gzipStream.Close()
        }

        if($null -ne $output){
            $output.Close()
        }

        if($null -ne $input){
            $input.Close()
        }
    }

    [xml]$primaryContent = Get-Content $outfile
    ClearTempFiles @($primaryFileGz, $outfile)
    return $primaryContent
}

function GetPackage {
    param(
        [string] $packageUrl,
        [string] $downloadDir,
        [System.Security.Cryptography.X509Certificates.X509Certificate2] $entitlementCert
    )

    $request = ([System.Net.HttpWebRequest]::Create($packageUrl))
    $request.ClientCertificates.Add($entitlementCert) | Out-Null
    $response = $null
    $downloadStream = $null
    $fileStream = $null

    try{
        Write-Host "Downloading package from: '$packageUrl'..." -ForegroundColor Green
        $response = $request.GetResponse()
        $bufferSize = 1024
        $buffer = New-Object Byte[] $bufferSize
        $fileName = [System.IO.Path]::GetFileName($response.ResponseUri.AbsolutePath)
        $downloadStream = $response.GetResponseStream()
        $fileStream = New-Object System.IO.FileStream([System.IO.Path]::Combine($downloadDir, $fileName), [System.IO.FileMode]::Create)

        while(($bytesRead = $downloadStream.Read($buffer, 0, $bufferSize)) -ne 0){
            $fileStream.Write($buffer, 0, $bytesRead)
        }

        $fileStream.Flush()
        Write-Host "Done!" -ForegroundColor Green
    } catch {
        $errorMessage = $_.Exception.Message
        throw "An error occurred during download: '$errorMessage'"
    } finally {
        if($null -ne $response){
            $response.Close()
        }

        if($null -ne $downloadStream){
            $downloadStream.Close()
        }

        if($null -ne $fileStream){
            $fileStream.Close()
        }
    }
}

function ClearTempFiles {
    param(
        [string []] $tempFiles
    )

    foreach($file in $tempFiles){
        Remove-Item $file -Force
    }
}
