#Constants
#####
$ErrorActionPreference = "Stop"
#####

#
#Public functions
#
<#
 .SYNOPSIS
 Deletes the download directory, and removes certificates.

 .DESCRIPTION
 Deletes the download directory and all of its contents. Removes the certificates
 from the certificate store if they exist.

 .PARAMETER downloadDir
 The directory containing your downloaded packages. (CAUTION) Whatever this is set
 to will be deleted.

 .PARAMETER rootCAPath
 The path to the file containing the root CA certificate. It will be used to match
 and delete its corresponding certificate in the certificate store.

 .EXAMPLE
 Clear-EntitlementData -downloadDir "C:\some\path" -rootCAPath "C:\root\ca.pfx"
 #>
function Clear-EntitlementData {
    [CmdletBinding(SupportsShouldProcess = $true)]
    param(
        [Parameter(Mandatory = $true)]
        [string] $downloadDir,
        [Parameter(Mandatory = $true)]
        [string] $rootCAPath
    )

    try {
        if ($PSCmdlet.ShouldProcess("Removing certificte authority '$rootCAPath' from the system")) {
            RemoveCertificateAuthority $rootCAPath
        }

        if (Test-Path $downloadDir) {
            if ($PSCmdlet.ShouldProcess("Deleting download directory '$downloadDir' from the system")) {
                Write-Verbose "Deleting contents of download directory: '$downloadDir'..."
                Remove-Item -Path $downloadDir -Recurse -Force
            }
        }
    }
    catch {
        throw
    }
}

<#
 .SYNOPSIS
 Downloads the specified package from a repo.

 .DESCRIPTION
 Imports the appropriate root and intermediate CA certificates,
 and the downloads the list of package passed in. If no packages
 are specified, all available packages will be downloaded.

 .PARAMETER packageNames
 An array of packages to download. If unspecified, takes all pacakges
 that are entitled.

 .PARAMETER downloadDir
 The directory in which to store the downloaded package. This directory
 will be created if it doesn't exist.

 .PARAMETER rootCAPath
 The path to the file containing the root CA certificate.

 .PARAMETER entitlementCertPath
 The path to the file containing the entitlement certificate.

 .PARAMETER baseUrl
 Url to the base of the package repository. Defaults to the SAS
 production repository if unspecified.

 .EXAMPLE
 Get-Entitlement -downloadDir "C:\some\path" -rootCAPath "C:\root\ca.pem" `
 -entitlementCertPath "C:\title\cert.pfx" -packageNames 'sas-ml','sas-stat'

 .EXAMPLE
 Get-Entitlement -downloadDir "C:\some\path" -rootCAPath "C:\root\ca.pem" `
 -entitlementCertPath "C:\title\cert.pfx"

 .EXAMPLE
 Get-Entitlement -downloadDir "C:\some\path" -rootCAPath "C:\root\ca.pem" `
 -entitlementCertPath "C:\title\cert.pfx" -packageName "sas-tkcore" `
 -baseUrl https://repoexample.sas.com/repos
 #>
function Get-Entitlement {
    [CmdletBinding()]
    param(
        [string[]] $packageNames,
        [Parameter(Mandatory = $true)]
        [string] $downloadDir,
        [Parameter(Mandatory = $true)]
        [string] $rootCAPath,
        [Parameter(Mandatory = $true)]
        [string] $entitlementCertPath,
        [string] $baseUrl = ''
    )

    try {
        New-Item -ItemType Directory -Force -Path $downloadDir | Out-Null
        $metadata = $null

        if ($baseUrl -eq '') {
            $metadata = @(Get-EntitlementMetadata -packageNames $packageNames -entitlementCertPath $entitlementCertPath `
                    -rootCAPath $rootCAPath)
        }
        else {
            $metadata = @(Get-EntitlementMetadata -packageNames $packageNames -entitlementCertPath $entitlementCertPath `
                    -rootCAPath $rootCAPath -baseUrl $baseUrl)
        }

        if ($null -eq $packageNames) {
            $packageNames = @($metadata.name)
        }

        foreach ($name in $packageNames) {
            $package = $metadata.Where{$_.name -eq $name}
            if ($null -ne $package) {
                Get-EntitlementByUrl -downloadDir $downloadDir -rootCAPath $rootCAPath `
                    -entitlementCertPath $entitlementCertPath -url $package.Url `
                    -checksum $package.Checksum -checksumType $package.ChecksumType
                continue
            }
            Write-Error "Package '$name' was not found in any repo."
        }
    }
    catch {
        throw
    }
}

<#
  .SYNOPSIS
  Downloads packages directly from a URL.

  .DESCRIPTION
  Takes known url for a package, and downloads it to
  a local directory. This cmdlet will most likely need to be used in
  combination with Get-EntitlementMetadata

  .PARAMETER downloadDir
 The directory in which to store the downloaded package. This directory
 will be created if it doesn't exist.

 .PARAMETER rootCAPath
 The path to the file containing the root CA certificate.

 .PARAMETER entitlementCertPath
 The path to the file containing the entitlement certificate.

 .PARAMETER url
 A URL to the package you want to download.

 .Example
Get-EntitlementByUrl -downloadDir 'C:\some\dir' -rootCAPath 'C:\root\cert.pfx' `
-entitlementCertPath 'C:\entitlement\cert.pfx' -url 'http://some/package.msi'
#>
function Get-EntitlementByUrl {
    [CmdletBinding()]
    param(
        [Parameter(Mandatory = $true)]
        [string] $downloadDir,
        [Parameter(Mandatory = $true)]
        [string] $rootCAPath,
        [Parameter(Mandatory = $true)]
        [string] $entitlementCertPath,
        [Parameter(Mandatory = $true)]
        [string] $url,
        [string] $checksum = '',
        [string] $checksumType = ''
    )

    try {
        ImportCertificateAuthority $rootCAPath
        New-Item -ItemType Directory -Force -Path $downloadDir | Out-Null
        $fileName = $url.Split("/")[-1]
        $pkgPath = Join-Path $downloadDir $fileName

        if (shouldDownloadPackage $pkgPath $checksum $checksumType) {
            FetchEntitlement $downloadDir $url $entitlementCertPath
            return
        }

        Write-Host "Package '$fileName' already present. Skipping download." -ForegroundColor Green
    }
    catch {
        throw
    }
}

<#
  .SYNOPSIS
  Fetches metadata for entitled packages

  .DESCRIPTION
  Given a list of packages, fetches the metadata for the packages
  in said list. If no packages are specified, the metadata for
  all packages that the user is entitled to will be returned.
  Metadata will only be returned for a single version (latest) of
  a package.

  .PARAMETER packageNames
  An array containing the package names to get metadata for.
  If none are specified, metadata on all pacakges will be
  returned. Duplicate entries are ignored.

 .PARAMETER rootCAPath
 The path to the file containing the root CA certificate.

 .PARAMETER entitlementCertPath
 The path to the file containing the entitlement certificate.

 .PARAMETER baseUrl
 Url to the base of a package repository. Defaults to the SAS
 production entitlement repo.

 .Example
Get-EntitlementMetadata -rootCAPath 'C:\root\cert.pfx' -entitlementCertPath 'C:\entitlement\cert.pfx'`
-packageNames 'sas-ml','sas-stat'

 .Example
Get-EntitlementMetadata -rootCAPath 'C:\root\cert.pfx' -entitlementCertPath 'C:\entitlement\cert.pfx'`
-baseUrl 'https://examplerepo.sas.com/repos'
#>
function Get-EntitlementMetadata {
    [CmdletBinding()]
    param(
        [string[]] $packageNames,
        [Parameter(Mandatory = $true)]
        [string] $entitlementCertPath,
        [Parameter(Mandatory = $true)]
        [string] $rootCAPath,
        [string] $baseUrl = "https://ses.sas.download/ses/"
    )

    try {
        $packageData = New-Object System.Collections.ArrayList

        if (!(Test-Path $entitlementCertPath)) {
            throw "Entitlement certificate '$entitlementCertPath' does not exist"
        }

        ImportCertificateAuthority $rootCAPath
        $entitlementCert = New-Object System.Security.Cryptography.X509Certificates.X509Certificate2($entitlementCertPath)
        Write-Host "Fetching repository metadata..."

        #Retrieve and process ODD data
        $entitlementData = GetEntitlementData $baseUrl $entitlementCert
        $repos = CompileRepoData $baseUrl $entitlementCert $entitlementData.Repositories
        $repoController = [RepoController]::new($repos, $entitlementCert)

        if ($null -eq $packageNames -or $packageNames.Count -eq 0) {
            $packageNames = $repoController.odd.GetProducts()
        }

        foreach ($name in $packageNames) {
            Write-Verbose "Getting data for package '$name'"
            $pkgData = $repoController.GetPackageData($name)

            if ($pkgData.Package -eq $null) {
                Write-Warning "Package is not found in any package repository for the product '$name'."
                continue
            }

            $metadata = New-Object -TypeName psobject -Property @{
                Url          = $pkgData.Url
                Name         = $pkgData.Package.name
                ProductCode  = $pkgData.Package.productCode
                UpgradeCode  = $pkgData.Package.upgradeCode
                ProductName  = $pkgData.Package.productName
                Checksum     = $pkgData.Package.checksum
                ChecksumType = $pkgData.Package.ChecksumType
                Version      = $pkgData.Package.version
                TwelveByte   = $pkgData.TwelveByte
            }
            $packageData.Add($metadata) | Out-Null
        }
    }
    catch {
        throw
    }
    Write-Host "Done!"

    if ($packageData.Count -eq 0) {
        throw "No package data could be found for any packages."
    }

    return $packageData.ToArray()
}

<#
.SYNOPSIS
Returns a list of updatable products

.DESCRIPTION
Given a name/version hashtable, returns a list of custom objects containing
those products that are elegible for upgrades and the versions to be
upgraded to.

.PARAMETER products
A hashtable in format @{"name" = "version";...}

.PARAMETER entitlementCertPath
Path to an entitlement certificate

.PARAMETER rootCAPath
Path to the SAS root CA

.EXAMPLE
Show-Update -products @{"coolproduct" = "1.2.3.4"} -entitlementCertPath "C:\ent\cert.pfx" -rootCAPath "C:\root\ca.pfx"

.NOTES
This function will request metadata every time it is called, so it's more efficient to build your hashtable of
products before calling this function.
#>

function Show-Update {
    [CmdletBinding()]
    param(
        [Parameter(Mandatory = $true)]
        [hashtable] $products, # hash of product/version pairs
        [Parameter(Mandatory = $true)]
        [string] $entitlementCertPath,
        [Parameter(Mandatory = $true)]
        [string] $rootCAPath,
        [switch] $checkAllVersions
    )

    $updates = New-Object System.Collections.ArrayList
    try {
        $metadata = @(Get-EntitlementMetadata -packageNames $products.Keys -entitlementCertPath $entitlementCertPath `
                -rootCAPath $rootCAPath)

        foreach ($entry in $products.GetEnumerator()) {
            Write-Host "Checking for new version of '$($entry.Name)'..."
            $pkg = $metadata.Where{$_.Name -eq $entry.Name } | Select-Object -First 1

            #if repo version is later. Add stuff here for comparison at level 3 or 4

            if (($null -ne $pkg) -and (CompareVersionData -first $pkg.Version -second $entry.Value -checkSasVersions:$checkAllVersions) -gt 0 ) {
                Write-Host "New version '$($pkg.Version)' found"
                $updatePkg = New-Object -TypeName psobject -Property @{
                    Name           = $entry.Name
                    CurrentVersion = $entry.Value
                    UpdateVersion  = $pkg.Version
                }
                $updates.Add($updatePkg) | Out-Null
                continue
            }

            Write-Host "No new version found"
        }
    }
    catch {
        throw
    }
    return $updates
}


<#
.SYNOPSIS
Surfaces the ability to compare verions

.DESCRIPTION
Given two versions, compare them and return 1 if first is later,
-1 if second is greater, and 0 if they're equal. By default, it will only
compare on the first three version fields unless the compareWholeVersion
switch is given.

.EXAMPLE
Compare-Version -first '1.2.3.4' -second '1.2.0.0' -compareWholeVersion
#>
function Compare-Version {
    [CmdletBinding()]
    param(
        [Parameter(Mandatory)]
        [string] $first,
        [Parameter(Mandatory)]
        [string] $second,
        [switch] $compareWholeVersion
    )

    return CompareVersionData -first $first -second $second -checkSasVersions:$compareWholeVersion
}

#
#Private functions
#
function JoinUrlPieces([string[]]$pieces) {
    #Sanitize all the pieces first
    for ($i = 0; $i -lt $pieces.Count; $i++) {
        $pieces[$i] = $pieces[$i].Trim().Trim('/')
    }

    $url = [string]::Join("/", $pieces)
    return $url
}

function FetchEntitlement {
    param(
        [Parameter(Mandatory = $true)]
        [string] $downloadDir,
        [Parameter(Mandatory = $true)]
        [string] $packageUrl,
        [Parameter(Mandatory = $true)]
        [string] $entitlementCertPath
    )

    if (!(Test-Path $entitlementCertPath)) {
        throw "Entitlement certificate '$entitlementCertPath' does not exist."
    }

    $entitlementCert = New-Object System.Security.Cryptography.X509Certificates.X509Certificate2($entitlementCertPath)
    GetPackage $packageUrl $downloadDir $entitlementCert
}