Set-Variable -Name installRegPath -Option Constant -Value "HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall"

<#
.SYNOPSIS
Install a sas MSI package.

.DESCRIPTION
Given a set of package metadata from SasEntitlement's Get-EntitlementMetadata function
Install all matching msi packages inside of the temporary directory provided.

.PARAMETER packageList
An array of metadata objects with the following properties:
New-Object -TypeName psobject -Property @{
    Url
    Name
    ProductCode
    UpgradeCode
    ProductName
    Checksum
    ChecksumType
    Version
    TwelveByte
}

.PARAMETER tempDir
Path to the directory containing the MSI packages to install

.PARAMETER installDir
Path to the directory to which the MSIs will be installed.

.PARAMETER installLogDir
Path to directory that will hold the logs for msiexec. Defaults
to $PSScriptRoot/logs

.EXAMPLE
$metadata = New-Object -TypeName psobject -Property @{
    Url          = 'www.someurl.com'
    Name         = 'mycoolpackage'
    ProductCode  = 'ITSAGUID'
    UpgradeCode  = 'YETANOTHERGUID'
    ProductName  = 'sas-product'
    Checksum     = '123098efeaab90482'
    ChecksumType = 'sha256'
    Version      = '1.2.3.4'
    TwelveByte   = 'sastb'
}
$pkgs = @($metadata)
Install-SasPackage -packageList $pkgs -tempDir C:\temp -installDir 'C:\Program Files\SAS'

.NOTES
This function is designed to be used in conjunction with SasEntitlement.
#>
function Install-SasPackage {
    [CmdletBinding()]
    param(
        [psobject[]] $packageList,
        [string] $downloadDir,
        [string] $installDir,
        [string] $installLogDir = $(Join-Path "$PSScriptRoot" "logs")
    )

    if (!(Test-Path $installLogDir)) {
        New-Item -ItemType Directory -Path $installLogDir | Out-Null
    }

    foreach ($package in $packageList) {
        $registryKey = Join-Path $installRegPath $package.ProductCode
        $msiFile = $package.Url.Substring($package.Url.LastIndexOf("/") + 1)
        $msiPath = Join-Path $downloadDir $msiFile
        $timestamp = Get-Date -Format FileDateTime
        $logPath = Join-Path "$installLogDir" "deploy-$($package.Name)-$timestamp.log"
        $datetime = Get-Date

        if (Test-Path $registryKey) {
            Write-Host "$($package.Name) is already installed on the system." -ForegroundColor Green
        }
        else {
            if (!(Test-Path $msiPath)) {
                throw "Package at '$msiPath' could not be found to install."
            }

            Write-Host "Task: Installing product $($package.Name) " -ForegroundColor Green -nonewline;
            Write-Host "started $datetime." -ForegroundColor Green
            Write-Host "This MSI log file is located at $logPath"
            $exitCode = (Start-Process -FilePath msiexec.exe "/i `"$msiPath`" /qn /norestart /L*V `"$logPath`" INSTALLDIR=`"$installDir`"" -Wait -Passthru).ExitCode
             if ($exitCode -eq 0) {
	       Write-Host "Install of this MSI was successful." -ForegroundColor Green
	       }
	      else {
	      Write-Warning "The $($package.Name) package did not install successfully. Consult the MSI log for further information about the error." 
	      Write-Warning "When the issue preventing the MSI from installing has been corrected, run setup.bat again to install any failed packages."		
	                }
            Write-Host "Task complete." -ForegroundColor Green

        }
    }
}

<#
.SYNOPSIS
Stores data to use at uninstall time.

.DESCRIPTION
Stores metadata about the installed packages so it isn't necessary to
hit the repository for metadata during uninstall. It also provides easy
lookup for what products are installed.

.PARAMETER packageList
An array of metadata objects with the following properties:
New-Object -TypeName psobject -Property @{
    Url
    Name
    ProductCode
    UpgradeCode
    ProductName
    Checksum
    ChecksumType
    Version
    TwelveByte
}

.PARAMETER uninstallDataFile
Path to the file which will store the uninstall data. It will be the
exported xml representation of a powershell object.

.EXAMPLE
metadata = New-Object -TypeName psobject -Property @{
    Url          = 'www.someurl.com'
    Name         = 'mycoolpackage'
    ProductCode  = 'ITSAGUID'
    UpgradeCode  = 'YETANOTHERGUID'
    ProductName  = 'sas-product'
    Checksum     = '123098efeaab90482'
    ChecksumType = 'sha256'
    Version      = '1.2.3.4'
    TwelveByte   = 'sastb'
}
$pkgs = @($metadata)
Save-UninstallData -packageList $pkgs -uninstallDataFile C:\my\uninstall\data.xml

.NOTES
Metadata for $packageList can be acquired from the SasEntitlement module using
Get-EntitlementMetadata
#>
function Save-UninstallData {
    [CmdletBinding()]
    param(
        [psobject[]] $packageList,
        [string] $uninstallDataFile
    )

    $uninstallData = Import-UninstallData $uninstallDataFile
    $newData = NewUninstallData $packageList
    $mergedData = MergeUninstallData $uninstallData $newData
    WriteUninstallData $uninstallDataFile $mergedData
}

function Uninstall-SasPackage {
    [CmdletBinding()]
    param(
        [string] $installDir,
        [string] $installLogDir = $(Join-Path "$PSScriptRoot" "logs"),
        [string] $uninstallDataFile
    )

    if (!(Test-Path $installLogDir)) {
        New-Item -ItemType Directory -Path $installLogDir | Out-Null
    }

    $uninstallData = Import-UninstallData $uninstallDataFile

    foreach ($entry in $uninstallData.GetEnumerator()) {
        $registryKey = Join-Path $installRegPath $entry.Value.productCode
        $logpath = Join-Path "$installLogDir" "$(($entry.Value).Name).log"

        if (-Not (Test-Path $registryKey)) {
            Write-Host "$(($entry.Value).Name) is not installed on the system." -ForegroundColor Green
        }
        else {
            Write-Host "Task: Uninstalling product $(($entry.Value).Name)." -ForegroundColor Green
            Start-Process -FilePath msiexec.exe "/x $(($entry.Value).productCode) /qn /norestart /L*V `"$logpath`"" -Wait
            Write-Host "Task complete." -ForegroundColor Green
        }
    }
}

<#
.SYNOPSIS
Copies the deployment modules onto the system.

.DESCRIPTION
This function will copy all of the deployment modules to the install directory
where the MSI packages are being laid down. The copy only includes files that are
part of the deployment tooling.

.PARAMETER sourceDir
Path to the source directory from which to copy. In most cases, it's the deployment directory
containing these modules.

.PARAMETER destinationDir
Path to the destination directory, which should be the same as the install directory for the
MSIs

.EXAMPLE
Install-DeploymentModules -sourceDir C:\downloads\SASViyaDeployment -destinationDir 'C:\Program Files\SAS'

.NOTES
General notes
#>
function Install-DeploymentModules {
    [CmdletBinding()]
    param(
        [string] $sourceDir,
        [string] $destinationDir
    )

    if ($sourceDir -ne $destinationDir) {
        Write-Verbose "Copying deployment modules to '$destinationDir'"
        New-Item -ItemType Directory -Path $destinationDir -Force | Out-Null
        $libDir = Join-Path $sourceDir "library"
        $configDir = Join-Path $sourceDir "config"

        #copy the library dir with modules
        Copy-Item -Path $libDir -Destination $destinationDir -Recurse -Force
        #copy config dir
        Copy-Item -Path $configDir -Destination $destinationDir -Recurse -Force
        #and all the deployment scripts...
        Copy-Item -Path "$sourceDir\*" -Destination $destinationDir -Include '*.ps1', '*.bat', '*.psd1' -Force

        return
    }

    Write-Verbose "Running deployment from installation directory. Skipping module copy."
}

<#
.SYNOPSIS
Loads uninstall data from a file into an object

.DESCRIPTION
Given the path to an xml file containing uninstall data, import that
data and format it into a hashtable.

.PARAMETER uninstallDataFile
Path to an xml file.

.EXAMPLE
Import-UninstallData C:\path\to\uninstalldata.xml
#>
function Import-UninstallData {
    [CmdletBinding()]
    param(
        [string] $uninstallDataFile
    )

    $uninstallData = @{}
    if (Test-Path $uninstallDataFile) {
        $uninstallData = Import-Clixml $uninstallDataFile
    }
    return $uninstallData
}


#return a custom object list of installed products
function Get-InstalledProducts {
    [CmdletBinding()]
    param ()

    $SasInstallRegPath = 'SOFTWARE\SAS Institute Inc.'
    $subKeyNames = GetSubKeysFromPath $SasInstallRegPath

    return ReadProductInfo $SasInstallRegPath $subKeyNames
}