#Copyright (c) 2017, SAS Institute Inc., Cary, NC, USA, All Rights Reserved

<#
.Synopsis
Imports/Exports a psd1 file using the rules established for a SAS service config file.

.Description
This module allows for the importing and exporting of psd1 files used for setting environment
and script variables.  This will change/add global scope variables and environment variables.

Access to script and environment variables is controlled by an access whitelist.  By default
the allowed variable names are hardcoded into this signed script.  This behavior can be
overridden by consumers of this module by providing positional argumnents to the
import-module command.

.Parameter 0
Can be an array of allowed environment variable names (wildcard characters are respected),
a string containing a wildcard pattern or $null.  If this parameter is not given or
is $null the default list is used.

.Parameter 1
Can be an array of allowed script variable names (wildcard characters are respected),
a string containing a wildcard pattern or $null.  If this parameter is not given or
is $null the default list is used.

.Example
import-module .\SasPsd1FileManagement

import-module .\SasPsd1FileManagement -args $null "*"

import-module .\SasPsd1FileManagement -args @( "_*", "sas*") "*"

#>

param(
    [parameter(Position=0, Mandatory=$false)] [Object]
        $allowedEnvVars = $null,
    [parameter(Position=1, Mandatory=$false)] [Object]
        $allowedScriptVars = $null
    )

Import-Module -force "$PSScriptRoot\..\Logging" `
              -ErrorAction "SilentlyContinue" `
              -ErrorVariable e
    if ($? -eq $false) {
        Write-Warning "Error importing module"
        Write-Warning $e[0]
        Write-Warning "trying again with verbose option"
        Import-Module -verbose -force "$PSScriptRoot\..\Logging"
        }

$AllowedEnvironmentVariables = @(
    "cas_port",
    "connect_args",
    "consul_bind_external",
    "consul_bootstrap_expect",
    "consul_datacenter_name",
    "consul_server_flag",
    "deployment_id",
    "entitlement_certificate",
    "instance_id",
    "java_bin",
    "java_global_option_cicompilercount",
    "java_global_option_consul_instanceid",
    "java_global_option_disableendpoints",
    "java_global_option_enablehealth",
    "java_global_option_enablemetrics",
    "java_global_option_javaagent",
    "java_global_option_jvmroute",
    "java_global_option_security_egd",
    "java_global_option_server_port",
    "java_global_option_tmpdir",
    "java_global_option_truststore",
    "java_global_option_xms",
    "java_global_option_xmx",
    "java_home",
    "license_filename",
    "package_repository_url",
    "public_config_dir",
    "sas_ca_certificate",
    "sashome_utils_libs_home",
    "spring_cloud_consul_host",
    "spring_cloud_consul_port",
    "temp_download_dir"
    )

$AllowedScriptVariables = @(
    "app_id",
    "app_name",
    "connect_args",
    "connect_cmd",
    "connect_port",
    "consul_config_dir",
    "consul_options",
    "consul_port_8500_tcp_add",
    "consul_url",
    "custom_ld_library_path",
    "display_name",
    "sascasappsrvenabled",
    "sascasdatadir",
    "sascasinstance",
    "sascontrollerhost",
    "sascontrollerhttpport",
    "sascontrollerport",
    "sasgroup",
    "sasinstance",
    "sas_java_home",
    "saspermstore",
    "sastenant",
    "sasuser",
    "sassfversion"
    "spring_cloud_consul_host",
    "spring_cloud_consul_port",
    "workspaceserver_log_path"
    )

if ($allowedEnvVars -is "Array") {
    $AllowedEnvironmentVariables = $allowedEnvVars
    }
elseif ($allowedEnvVars -is "String") {
    $AllowedEnvironmentVariables = @( $allowedEnvVars )
    }

if ($allowedScriptVars -is "Array") {
    $AllowedScriptVariables = $allowedScriptVars
    }
elseif ($allowedScriptVars -is "String") {
    $AllowedScriptVariables = @( $allowedScriptVars )
    }


#------------------------------------------------------------------------------------------
function ProcessSasPsd1Data([Hashtable] $data) {
    $rv = $true

    # Step 1
    if ($data.ContainsKey("Psd1IncludeFiles")) {
        $files = @()
        if ($data.Psd1IncludeFiles -is "Array") {
            $files = $data.Psd1IncludeFiles
            }
        elseif ($data.Psd1IncludeFiles -is "String") {
            $files = @( $data.Psd1IncludeFiles )
            }
        else {
            Write-LogWarning "Psd1IncludeFiles entry found, but content isn't a string or array (ignoring)"
            }
        foreach ($f in $files) {
            Write-LogDebug ("File: '{0}'" -f $f)
            if ($f -isnot "String") {
                Write-LogWarning ("Psd1IncludeFiles entry '{0}' isn't a string type (ignoring)" -f $f)
                continue
                }
            if ($f.ToLower().EndsWith("psd1") -eq $false) {
                Write-LogWarning ("Psd1IncludeFiles entry '{0}' doesn't have a PSD1 extension (ignoring)" -f $f)
                continue
                }
            if (Test-Path($f)) {
                Write-LogDebug ("Including file: '{0}'" -f $f)
                $result = Import-SasPsd1($f)
                if ($result -eq $false) {
                    Write-LogWarning ("Import of include file: '{0}' failed" -f $f)
                    }
                }
            else {
                Write-LogWarning ("Unable to include file: '{0}', file not found" -f $f)
                }
            }
        $data.Remove("Psd1IncludeFiles")
        }

    # Step 2
    $environment = @{}
    if ($data.ContainsKey("Environment") -and
        $data.Environment -is "Hashtable") {
        $environment = $data.Environment
        $data.Remove("Environment")
        }

    $switch = @{}
    $switchRx = "(?<name>[a-zA-Z_]+)_[Bb][Yy](?<byvar>[a-zA-Z_]+)"
    $rxOpts = [System.Text.RegularExpressions.RegexOptions]::None
    $timeout = [TimeSpan]::FromSeconds(1.0)
    $deleteKeys = @()
    foreach ($k in $data.Keys) {
        try {
            $matches = [RegEx]::Matches($k, $switchRx, $rxOpts, $timeout)
            if ($matches.Count -gt 0) {
                $rec = @{}
                $rec.Name = $matches[0].Groups["name"].Value
                $rec.ByVar = $matches[0].Groups["byvar"].Value
                $rec.Content = $data[$k]
                $switch.Add($k, $rec)
                $deleteKeys += $k
                }
            if ($matches.Count -gt 1) {
                Write-LogWarning ("Unexpected match for '{0}'" -f $k)
                }
            }
        catch [System.Text.RegularExpressions.RegexMatchTimeoutException] {
            Write-LogWarning ("Entry '{0}' generated timeout exception '{1}'" -f $k, $_.Exception.Message)
            continue;
            }
        }

    # delete already processed keys
    foreach ($k in $deleteKeys) {
        $data.Remove($k)
        Write-LogDebug ("Removing already processed key '{0}'" -f $k)
        }

    foreach ($k in $data.Keys) {
        if ($AllowedScriptVariables -like $k) {
            Write-LogDebug ("Setting: {0} = {1}" -f $k, $data[$k])
            Set-Variable -Name $k -Value $data[$k] -Scope "Global"
            }
        else {
            Write-LogWarning ("Unable to set: '{0} = {1}', name not allowed" -f $k, $data[$k])
            }
        }

    # Step 3
    foreach ($k in $switch.Keys) {
        Write-LogDebug ("Switch {0} by {1}" -f $k, $switch[$k].ByVar)
        $name = $switch[$k].Name
        $byVar = $switch[$k].ByVar

        if ($AllowedScriptVariables -like $name) {
            if ($AllowedScriptVariables -like $byVar) {
                $index = Get-Variable -Name $byVar -ValueOnly
                $values = $switch[$k].Content
                if ($switch[$k].Content -is "Array") {
                    if ($index.GetType().Name -match 'byte|int16|int32|int64|sbyte|uint16|uint32|uint64') {
                        }
                    else {
                        Write-LogWarning ("Unable to set: '{0} by {1}', by variable is not a valid numeric index type" -f $name, $byVar)
                        continue
                        }
                    }
                elseif ($switch[$k].Content -is "Hashtable") {
                    if ($values.ContainsKey($index)) {
                        $v = $values[$index]
                        if ($v -is "Hashtable") {
                            Process-SasPsd1Data $v
                            continue
                            }
                        }
                    else {
                        Write-LogWarning ("Unable to set: '{0} by {1}', content does not contain by value" -f $name, $byVar)
                        continue
                        }
                    }
                else {
                    Write-LogWarning ("Unable to set: '{0} by {1}', content is not index-able" -f $name, $byVar)
                    continue
                    }

                Write-LogDebug ("Setting {0} = '{1}'" -f $name, $values[$index])
                Set-Variable -Name "$name" -Value $values[$index] -Scope "Global"
                }
            else {
                Write-LogWarning ("Unable to set: '{0} by {1}', by name not allowed" -f $k, $switch[$k].ByVar)
                }
            }
        else {
            Write-LogWarning ("Unable to set: '{0}', name not allowed" -f $name)
            }
        }

    # Step 4
    foreach ($k in $environment.Keys) {
        if ($AllowedEnvironmentVariables -like $k) {
            Write-LogDebug ("Set env:{0} = '{1}'" -f $k, $environment[$k])
            [Environment]::SetEnvironmentVariable($k, $environment[$k])
            }
        else {
            Write-LogWarning ("Unable to set env:{0} = '{1}', name not allowed" -f $k, $environment[$k])
            }
        }
    return $rv
    }


#------------------------------------------------------------------------------------------
function ArrayToString([Array] $array,
                        [System.Text.StringBuilder] $sb,
                        [String] $indent = "") {
    $sb.AppendLine($indent + "@(");
    $pad = $indent + "    "

    for ($i = 0; $i -lt $array.Length; $i++) {
        if ($array[$i] -ne $null) {
            if ($array[$i].GetType().IsValueType) {
                $sb.AppendLine(("{0}{1}" -f $pad, $array[$i]))
                }
            elseif ($array[$i] -is "String") {
                $sb.AppendLine(('{0}"{1}"' -f $pad, $array[$i].Replace('"',"`"")))
                }
            elseif ($array[$i] -is "Array") {
                ArrayToString $array[$i] $sb $pad
                }
            elseif ($array[$i] -is "Hashtable") {
                HashToString $array[$i] $sb $pad
                }
            else {
                Write-LogWarning ("Unexpected element data type {0} for variable {1}" -f $array[$i].GetType().Name, $k)
                }
            }
        }
    $sb.AppendLine($indent + ")");
    }


#------------------------------------------------------------------------------------------
function HashToString([Hashtable] $hash,
                        [System.Text.StringBuilder] $sb,
                        [String] $indent = "") {
    $sb.AppendLine($indent + "@{");
    $pad = $indent + "    "

    foreach ($k in ($hash.keys|sort-object)) {
        $name = $k
        $value = $hash[$k]

        if ($value -eq $null) {
            continue
            }

        if ($value.GetType().IsValueType) {
            $sb.AppendLine(("{0}{1} = {2}" -f $pad, $name, $value))
            continue
            }
        elseif ($value -is "String") {
            $sb.AppendLine(('{0}{1} = "{2}"' -f $pad, $name, $value.Replace('"',"`"")))
            continue
            }
        elseif ($value -is "Array") {
            $sb.Append(("{0}{1} = " -f $pad, $name))
            ArrayToString $value $sb $pad
            }
        elseif ($value -is "Hashtable") {
            $sb.Append(("{0}{1} = " -f $pad, $name))
            HashToString $value $sb $pad
            continue
            }
        else {
            Write-LogWarning ("Unexpected data type {0} for variable {1}" -f $vtype.Name, $k)
            }
        }

    $sb.AppendLine($indent + "}");
    }



<#
.Synopsis
Exports a psd1 file based on a given hashtable.

.Description
This function exports a psd1 file by "unrolling" a hashtable.  There
is no whitelist checking and the type of each (non-array, non-hashtable)
element must be able to serialize and deserialize using simple assignment
types.  It should be noted that when reading a psd1 file order can be important

.Parameter saspsd1
Hashtable of values to write into the psd1file

.Parameter psd1file
Path to the psd1 file to write.

.Parameter forceOverwrite
If true and a file exists with the name given by psd1file exists, the file will be
overwritten.

.Outputs
returns True on success, False on failure
#>
function Export-SasPsd1 {

    [CmdletBinding()]
    param ([Hashtable] $saspsd1,
           [string] $psd1file,
           [Boolean] $forceOverwrite = $false)

    $rv = $true
    try {
        Write-LogDebug ("Exporting: '{0}'" -f $psd1file)

        if ($saspsd1 -isnot "Hashtable") {
            Write-LogError("Incorrect variable type given for saspsd1")
            return $false
            }

        $sb = New-Object System.Text.StringBuilder
        HashToString $saspsd1 $sb

        if ($forceOverwrite -eq $false) {
            $rv = Test-Path -Path $psd1file
            if ($rv -eq $true) {
                Write-LogError ("File: '{0}' exists, but forceOverwrite not requested" -f $psd1file) `
                                -Category  ([System.Management.Automation.ErrorCategory]::OpenError)
                return $rv
                }
            $rv = $true
            }

        if ($rv -eq $true) {
            $sb.ToString() | Out-File $psd1file
            Write-LogDebug ("File '{0}' written" -f $psd1file)
            }
        }
    catch {
        Write-LogError -ErrorRecord $_
        $rv = $false
        }
    return $rv
    }


<#
.Synopsis
Imports a psd1 file using the rules established for a Sas service config file.

.Description
This function imports a psd1 file and processes it's content using
rules established for a Sas service config file.  This can change/add global
scope variables and environment variables.

.Parameter psd1file
Path to the psd1 file to process.

.Outputs
returns True on success, False on failure
#>
function Import-SasPsd1 {
    [CmdletBinding()]
    param ([string] $psd1file)

    $rv = $true
    try {
        Write-LogDebug ("Importing: '{0}'" -f $psd1file)
        $rv = Test-Path -Path $psd1file
        if ($rv -eq $false) {
            Write-LogError ("File: '{0}' not found" -f $psd1file) `
                            -Category  ([System.Management.Automation.ErrorCategory]::OpenError)
            return $rv
            }

        $data = Import-PowerShellDataFile $psd1file -ErrorAction "SilentlyContinue" `
                                                    -ErrorVariable e
        if ($?) {
            Write-LogDebug ("File: '{0}' read" -f $psd1file)
            $rv = ProcessSasPsd1Data $data
            }
        else {
            Write-LogError -ErrorRecord $e[0]
            $rv = $false
            }
        }
    catch {
        Write-LogError -ErrorRecord $_
        $rv = $false
        }
    return $rv
    }
